
# standard issue modules
from Globals import DTMLFile, MessageDialog, InitializeClass
from AccessControl import ClassSecurityInfo
from Acquisition import aq_parent, aq_base
import re
# from OFS.FindSupport import FindSupport
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from OFS.ObjectManager import ObjectManager
from OFS.Folder import Folder

#non-standard Zope requisite:
from Products.OrderedFolder import OrderedObjectManager, SubObjectTypeSupport

# my own

import Alias, Redirect, IHostOverride, IHost, utils

class Host(OrderedObjectManager.OrderedObjectManager, Folder):

    security = ClassSecurityInfo()
    security.declareObjectPublic()

    meta_type="Virtual Host"
    
    def possible_meta_types(self):
        return ObjectManager.all_meta_types(self, (IHostOverride.IHostOverride,))
    # icon="misc_/VirtualHostFolder/host"

    __implements__= IHostOverride.IAliasHostOverrideContainer, \
                    IHost.IHost, \
                    IHost.IModifiableHost, \
                    IHost.IModifiableRegexHost, \
                    IHost.ICMFSkinGetterHost, \
                    IHost.ICMFSkinSetterHost # actually includes getter but hi-ho for explicitness

    manage_options=(
        (
        {'label':'Configuration', 'action':'manage_configure',
        'help':('VirtualHostFolder','Host_Configure.stx')},
        {'label':'Contents', 'action':'manage_main',
         'help':('VirtualHostFolder','Host_Contents.stx')},
        {'label':'CMF Settings', 'action':'manage_cmf',
        'help':('VirtualHostFolder','VHF_CMF')}
        )
        + PropertyManager.manage_options
        + SubObjectTypeSupport.SubObjectTypeSupport.manage_options
        + SimpleItem.manage_options
        # + FindSupport.manage_options
        )


    #_properties=(
    #    {'id':'title', 'type': 'string', 'mode': 'w'},
    #    )

    # work with the domain list is always case insensitive
    # work with the domainRE is case insensitive by default but modifiable
    ## and now we need UI methods

    security.declareProtected( 'View management screens', 'manage_configure' )
    manage_configure = DTMLFile('dtml/HostManage_main', globals())

    security.declareProtected( 'Manage properties', 'manage_host' )
    def manage_host(self, domainList=[], virtualRoot="", domainRE="", domainREIgnoreCase=0, virtualRootTemplate="", REQUEST=None):
        """the main management function.  The mfunc uses the optional
        messaging system: if a callable object is passed to the various
        worker-bee functions, they will call the function with messaging
        information.  In this case, we are having the worker-bee functions
        append to a messages array that we then return to the caller."""
        messages=[]
        mfunc=messages.append
        self.delDomainsFromList([x for x in self.getDomainList() if x not in domainList], mfunc)
        self.addDomainsToList([x for x in domainList if x not in self.getDomainList()], mfunc)
        if virtualRoot != self.getVirtualRootURL():
            self.setVirtualRootURL(virtualRoot, mfunc)
        if domainRE!=self.getDomainREComponents()[0] or ((domainREIgnoreCase and 1 or 0) ^ (self.getDomainREComponents()[1] and 1 or 0)):
            self.setDomainRE(domainRE, ignorecase=domainREIgnoreCase, mfunc=mfunc)
        if virtualRootTemplate != self.getVirtualRootTemplate():
            self.setVirtualRootTemplate(virtualRootTemplate, mfunc)
        if REQUEST is not None:
            return self.manage_configure(self, REQUEST, messages=messages)
        return messages


    security.declareProtected( 'View management screens', 'manage_cmf_form' )
    manage_cmf_form = DTMLFile('dtml/HostManage_cmf', globals())

    security.declareProtected( 'Manage properties', 'manage_cmf' )
    def manage_cmf(self, managerList=[], ownerList=[], defaultSkin=None, REQUEST=None):
        """management function for cmf-related items.  Currently only skin
        stuff"""
        if REQUEST and not(managerList or ownerList or defaultSkin):
            return self.manage_cmf_form(self, REQUEST)
        msg=[]
        mfunc=msg.append
        self.setCMFManagerSkins(managerList,mfunc)
        self.setCMFOwnerSkins(ownerList,mfunc)
        self.setCMFDefaultSkin(defaultSkin,mfunc)
        if REQUEST is not None:
            return self.manage_cmf_form(self, REQUEST, messages=msg)
        return msg

    ## the non-UI basic methods

    # _aliasList=() # added for backwards compatibility
    def __init__(self):
        """yup, it's an init."""
        self._virtualRoot="/"
        self._domainList=()
        self._virtualRootTemplate=""
        self._domainRE=None
        self.CMF_manager_skins=[]
        self.CMF_owner_skins=[]
        self.CMF_default_skin=None

    # backwards compatibility
    CMF_manager_skins=CMF_owner_skins=CMF_default_skin=None

        ## the setters and getters.

    security.declareProtected( 'Manage properties', 'setVirtualRootURL' )
    def setVirtualRootURL(self, virtualRoot="/", mfunc=None):
        """sets the static virtual root.  Uses optional message system (mfunc)."""
        if not virtualRoot: virtualRoot="/"
        virtualRoot=utils.cleanPath.match(virtualRoot).group(1) # eliminates final slash if present
        self._virtualRoot = self._check_path(virtualRoot, mfunc) and virtualRoot or self._virtualRoot
        self._p_changed = 1
        if mfunc and self._virtualRoot==virtualRoot:
            if self._virtualRoot=="/":
                mfunc("Virtual host root set to Zope root")
            else:
                mfunc("Virtual host root changed")
        return self._virtualRoot==virtualRoot

    security.declarePublic( 'getVirtualRootURL' )
    def getVirtualRootURL(self):
        """get the static path."""
        return self._virtualRoot

    security.declarePublic( 'getVirtualRootURLWithRegex' ) # magic
    def getVirtualRootURLWithRegex(self, matchObj=None, matchObj2=None):
        """Call it with no arguments, and you will get the static
        path.  Call it with a regular expression match object or two
        (designed for given case vs. lowercase), and it will try to
        return the expanded (filled-in) version of the dynamic virtual
        root template; if the template does not exist or the expanded
        path does not resolve to a valid Zope folder, the static path
        is returned."""

        if matchObj and self._virtualRootTemplate:
            dynamicRoot=matchObj.expand(self._virtualRootTemplate)
            if self._check_path(dynamicRoot): return dynamicRoot
            if matchObj2: # experimental lowercase
                dynamicRoot=matchObj2.expand(self._virtualRootTemplate)
                if self._check_path(dynamicRoot): return dynamicRoot
        return self._virtualRoot

    security.declareProtected( 'Manage properties', 'addDomainsToList' )
    def addDomainsToList(self, domains, mfunc=None):
        """adds domains to list; uses optional message system"""
        domains=self._clean_domainList(domains) # shape into a lowercase tuple, remove blanks
        ln=len(self._domainList)
        self._domainList+=tuple([x for x in domains if x not in self._domainList and self._check_domain(x, mfunc)])
        ln=len(self._domainList)-ln
        if ln:
            self._p_changed = 1
            if mfunc:mfunc("%d domain%s added to domain list" % (ln, ln!=1 and "s" or ""))
        return ln

    security.declareProtected( 'Manage properties', 'delDomainsFromList' )
    def delDomainsFromList(self, domains, mfunc=None):
        """deletes domains from list; uses optional message system"""
        domains=self._clean_domainList(domains) # shape into a lowercase tuple
        ln=len(self._domainList)
        self._domainList=tuple([x for x in self._domainList if x not in domains])
        ln=ln-len(self._domainList)
        if ln:
            self._p_changed = 1
            if mfunc:mfunc("%d domain%s deleted from domain list" % (ln, ln!=1 and "s" or ""))
        return ln

    security.declarePublic( 'getDomainList' )
    def getDomainList(self):
        """returns an unsorted domain list"""
        return self._domainList

    security.declarePublic( 'getSortedDomainList' )
    def getSortedDomainList(self):
        """returns a sorted version of the domain list"""
        x=list(self._domainList)
        x.sort()
        return x

    security.declareProtected( 'Manage properties', 'setVirtualRootTemplate' )
    def setVirtualRootTemplate(self, tmpl, mfunc=None):
        """sets the virtual root template; supports the optional
        messaging system."""
        if not tmpl:
            if mfunc:mfunc("Virtual root template deleted.")
            self._virtualRootTemplate=""
            self._p_changed = 1
            return 1
        virtualRoot=utils.cleanPath.match(tmpl).group(1) # eliminates final slash if present
        if self.isUsefulTemplate.search(tmpl):
            self._virtualRootTemplate=tmpl
            if mfunc:mfunc("Virtual root template changed")
            self._p_changed = 1
            return 1
        if mfunc:mfunc(r"Virtual rule template not set: the template must use back references such as \1 or else it is not a dynamic template.  Just set the standard virtual root instead.  If you want the regular expression to go to a different standard virtual root, consider creating a new host.")
        return 0

    security.declarePublic( 'getVirtualRootTemplate' )
    def getVirtualRootTemplate(self):
        """returns the unprocessed vitual root template"""
        return self._virtualRootTemplate

    security.declareProtected( 'Manage properties', 'setDomainRE' )
    def setDomainRE(self, domainRE, ignorecase=1, mfunc=None):
        """sets the domain regular expression.  Pass in anything that evaluates
        to Boolean true in ignorecase to set it to true.  No other flags are
        available (although they could be added, I didn't feel they would be
        useful).  The messages array is an optional human-readable
        messaging system described in the doc string for _check_path below: feel
        free to ignore."""
        if domainRE:
            test=0
            try:
                test=re.compile(domainRE, ignorecase and re.I or 0)
            except (SyntaxError, re.error), e:
                # get_transaction().abort()
                if mfunc:mfunc("The domain regular expression could not be changed. The most likely error is a syntax error in your regular expression.  Python reports the following: %s" % "; ".join(utils.getErrorText(e)))
                return 0
            if mfunc:mfunc("The domain regular expression is %s" % (self._domainRE and "changed" or "added"))
            self._domainRE=test
        else:
            self._domainRE=None
            if mfunc:mfunc("The domain regular expression is deleted")
        self._p_changed = 1
        return 1

    security.declarePublic( 'getDomainRE' )
    def getDomainRE(self):
        """returns the current domain regular expression *object*"""
        return self._domainRE

    security.declareProtected( 'Manage properties', 'getDomainREComponents' )
    def getDomainREComponents(self):
        """returns the current domain regular expression and the ignorecase
        flag setting.  If no current domain regular expression is set, returns an empty string and the default 'on' ignorecase setting."""
        return self._domainRE and (self._domainRE.pattern, self._domainRE.flags and 1) or ("", 1)

    security.declarePrivate('_matchDomainRE') # unnecessary
    def _matchDomainRE(self, domain):
        """silently handles absence of RE; included in this group because it
        directly accesses the _domainRE rather than using the getter.  Perhaps
        a weird, misguided combination of anal and faux-speed-conscious, at
        worst."""
        return self._domainRE and self._domainRE.match(domain)

    security.declarePrivate( '_check_path' ) # currently unnecessary
    def _check_path(self,path,mfunc=None):
        """pass in a path, have it validated as bad (empty, not found, or not
        a folder) (0) or ok (1).  The optional mfunc thing is a trick to let the method return a clear on-off numeric success message, while allowing it also to communicate some human-readable messages should we want them.  Just assign a variable in your local scope to an array, pass it along with the path, and then the next time you look at it, through the magic of passing by reference, it might have a message from this function.  This particular function will currently only give you a message if it encounters a bad path."""
        if not path:
           if mfunc:mfunc("Empty path ignored" )
           return 0
        if path.endswith('/'):
           if len(path) == 1:
              return 1
           else:
              path = path[:-1]
        try:
            leaf_object = self.unrestrictedTraverse(path)
        except:
            if mfunc:mfunc("The path %s was not found, so ignored" % path)
            return 0
        if not getattr(leaf_object.aq_base,'isAnObjectManager', None):
            if mfunc:mfunc( "Paths must end in a container: %s does not." % path )
        return 1


    # the CMF stuff

    security.declarePublic('getCMFAllSkins')
    def getCMFAllSkins(self):
        """If the host basic virtual root is in a CMF, will return a list of
        the CMF skins; if not, returns an empty list.  This is a
        convenience function for use with the properties needed for the Virtual
        Host Tool"""
        try:
            skinTool=self.unrestrictedTraverse(self.getVirtualRootURL()+"/portal_skins")
            return skinTool.getSkinSelections()
        except:
            return []

    security.declarePublic('getCMFManagerSkins')
    def getCMFManagerSkins(self):
        """get the possible owner choices for skins as set by the manager"""
        prop= self.CMF_manager_skins
        allSkins=self.getCMFAllSkins()
        if not prop or not allSkins:
            return allSkins
        ret= [x for x in prop if x in allSkins]
        return ret or allSkins

    security.declarePublic('getCMFOwnerSkins')
    def getCMFOwnerSkins(self):
        """get the possible user choices for a skin, as selected by the
        site owner"""
        prop=self.CMF_owner_skins
        manager=self.getCMFManagerSkins()
        if not prop or not manager:
            return manager
        ret= [x for x in prop if x in manager]
        return ret or manager

    security.declarePublic('getCMFDefaultSkin')
    def getCMFDefaultSkin(self):
        """get the current default CMF skin"""
        prop= self.CMF_default_skin
        own=self.getCMFOwnerSkins()
        if not own:
            return None
        if prop and prop in own:
            return prop
        try:
            result=self.unrestrictedTraverse(self.getVirtualRootURL()+"/portal_skins").getDefaultSkin()
            if result in own: return result
        except:
            pass # we'll deal with it below
        return own[0]

    security.declareProtected( 'Manage properties', 'setCMFManagerSkins' )
    def setCMFManagerSkins(self, valSeq, mfunc=None):
        """set possible owner choices acording to manager"""
        allSkins=self.getCMFAllSkins()
        setSeq=[x for x in valSeq if x in allSkins]
        if not setSeq:
            if mfunc:
                mfunc("No valid skin names entered so manager list reset to full skin list")
            setSeq=allSkins
        self.CMF_manager_skins=setSeq
        return 1

    security.declareProtected( 'Manage properties', 'setCMFOwnerSkins' )
    def setCMFOwnerSkins(self, valSeq, mfunc=None):
        """set possible user choices according to the owner"""
        managerSkins=self.getCMFManagerSkins()
        setSeq=[x for x in valSeq if x in managerSkins]
        if not setSeq:
            if mfunc:
                mfunc("No valid skin names entered so owner list reset to full manager skin list")
            setSeq=managerSkins
        self.CMF_owner_skins=setSeq
        return 1

    security.declareProtected( 'Manage properties', 'setCMFDefaultSkin' )
    def setCMFDefaultSkin(self, skinName, mfunc=None):
        """set default CMF skin for this host"""
        ownerSkins=self.getCMFOwnerSkins()
        if skinName in ownerSkins:
            self.CMF_default_skin=skinName
            return 1
        if self.CMF_default_skin in ownerSkins:
            if mfunc:
                mfunc("No change to default skin: selection is not valid")
            return 0
        if ownerSkins:
            self.CMF_default_skin=ownerSkins[0]
            if mfunc:mfunc("Not a valid default skin: default skin set to %s" % self.CMF_default_skin)
            return 0
        self.CMF_default_skin=None
        if  mfunc: mfunc("No valid default skins: default skin set to None")
        return 0

    ## the other basic method(s)
    security.declarePublic( 'testActivate' )
    def testActivate(self, host, stack, VHFid, request):
        host=host.strip()
        lowhost=host.lower()
        if lowhost in self.getDomainList(): return (None, None)
        res=self._matchDomainRE(host)
        if res:
            res2= (host != lowhost) and self.getVirtualRootTemplate() and self._matchDomainRE(lowhost)
            return (res, res2)
        return None

    security.declarePublic( 'activateOverride' )
    def activate(self, res, host, stack, VHFid, request):
        res, res2=res
        path=self.getVirtualRootURLWithRegex(res, res2)
        request.set( "CANONICAL_VIRTUAL_HOST_ROOT_PATH", path )
        overrideList=self.objectValues()
        if overrideList:
            stackString=""
            for loc in stack: # bad.  I do this someplace else too.  I wonder just how inefficient it is...
                stackString="/"+loc+stackString
            for override in overrideList:
                resp= override.testActivate(host, stack, stackString)
                if resp:
                    resp= override.activate(resp, stack, VHFid, request, self.getVirtualRootURL(), [self.getId()]) # always use standard virtual root (not expanded) when dealing with aliases
                    if resp:
                        if type(resp)==type(''):
                            stackString=resp
                        elif resp==-1: break
                        else: return resp
        # set stack with no alias
        # request.set( "VIRTUAL_HOST_ROOT_PATH", path )
        request.set( "VIRTUAL_HOST_IDS", self.getId() )
        list_path = path.split('/')[1:]
        stack.append('/') # the flag that bobo_traverse looks for
        stack.append(VHFid)
        for i in range(len(list_path)):
            stack.append(list_path.pop())
        return 1


    security.declareProtected( 'Manage properties', 'checkParentsAndAliasName' )
    def checkParentsAndAliasName(self, parent, aliasName, caller, mfunc=None):
        aliasList=self.objectValues()
        for aliasObj in aliasList:
            if aliasObj != caller:
                objParent=aliasObj.getParent()
                if objParent==parent:
                    objName=aliasObj.getAliasName()
                    if objName==aliasName:
                        if mfunc: mfunc("An alias with the identical address (parent and name) already exists in the parent folder.")
                        return 0
                elif parent.startswith(objParent+"/"): #objParent+"/" == parent[:(len(objParent)+1)]:
                    if mfunc: mfunc("You appear to be trying to place an alias within another alias.  To do this, actually add an alias within the contents of the parent alias folder, rather than specifying it in the parent field.")
                    return 0
                elif objParent.startswith(parent+"/"): # parent+"/" == objParent[:(len(parent)+1)]:
                    if mfunc: mfunc("An already existing alias, %s, would be masked by this one if you used this parent.  If the other alias needs to exist within this one, you may need to delete it, change the parent in this alias to what you want, and then add the existing alias within this one.""" % aliasObj.getId())
                    return 0
        fullPath=self.getVirtualRootURL()+parent
        try:
            parentObj=self.unrestrictedTraverse(fullPath)
        except:
            if mfunc: mfunc("The parent must exist within the standard virtual root.  %s does not exist." % fullPath)
            return 0
        return 1

    security.declarePrivate('_check_domain') # unnecessary
    def _check_domain(self, domain, mfunc=None):
        """This is a plug to put in code that rejects domains.  At the
        moment, all it does is try to make sure the domain is not a numeric IP
        address.  That's probably all it will ever need to do too, but hey, here
        it is."""
        if utils.isIPAddress.match(domain):
            if mfunc:mfunc("%s : IP addresses are not mappable" % domain)
            return 0
        return 1

    security.declarePrivate('_clean_domainList') # unnecessary
    def _clean_domainList(self, domains, mfunc=None):
        """a utility function.  converts all of the domain names to lower
        case, which is convenient for the domain list functionality, and does
        whatever else needs doing to the list to shape it up."""
        return [x.lower().strip() for x in domains if x and not x.isspace() ]

    ################### API (some originally in VHF)

    security.declarePublic( 'getCanonicalVirtualRootURL' )
    def getCanonicalVirtualRootURL(self, REQUEST=None): # convenience; modified from VHF; NOT PART OF INTERFACE
        """Get the actual virtual host root URL (string); it will be
        the same as the above, unless overwritten by a HostOverride
        object such as an alias."""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        if REQUEST and REQUEST.has_key("CANONICAL_VIRTUAL_HOST_ROOT_PATH"):
            return REQUEST["CANONICAL_VIRTUAL_HOST_ROOT_PATH"]
        return self.getVirtualRootURL()

    security.declarePrivate( 'getCanonicalVirtualRootObject' )
    def getCanonicalVirtualRootObject(self, REQUEST=None): # convenience; modified from VHF; NOT PART OF INTERFACE
        """Get the canonical virtual host root object"""
        try:
            return self.unrestrictedTraverse(self.getCanonicalVirtualRootURL(REQUEST))
        except:
            return self.getVirtualRootObject()

    security.declarePrivate( 'getCanonicalVirtualRootObject' )
    def getVirtualRootObject(self):
        """Get the static virtual root object"""
        try:
            return self.unrestrictedTraverse(self.getVirtualRootURL())
        except:
            return None

    security.declarePublic( 'getVirtualURLToObject' )
    def getVirtualURLToObject(self, obj, REQUEST=None):
        """Pass in an object and get a URL (string) that should reach
        it from within the current virtual host"""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        try:
            path=self.getVirtualPathToObject(obj, REQUEST, 1)
            if path:
                url ="/".join(path)
                if not url or not url.startswith("/"):
                    url="/"+url
                return url
            return "/"
        except "Redirect", e:
            return e


    security.declarePublic( 'getVirtualPathToObject' )
    def getVirtualPathToObject(self, obj, REQUEST=None, raiseRedirect=0): #  ?
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
#        if not hasattr(obj, "getPhysicalPath"):
#            return []
        objPath=obj.getPhysicalPath()
        #if not self.getVirtualHostStatus(REQUEST): return objPath
        rootObj=self.getCanonicalVirtualRootObject(REQUEST)
        rootPath=rootObj.getPhysicalPath()
        hostObj=self
        domainHost=REQUEST['HTTP_HOST'].split(':')[0] ##### add host into API?
        over=res=ct=0 # declaring "over" isn't necessary in python, in which scope is just either global or local; however, I think it is clear and a good habit
        lenR=len(rootPath)
        lenO=len(objPath)
        while lenR>ct and lenO>ct and rootPath[ct]==objPath[ct]:
            ct+=1
        overrides=hostObj.objectValues() # so these are all different kinds of overrides
        if lenR==ct: # we have a match--as long as an override is not masking the path, we can send along objPath[ct:]
            if not overrides or lenO==ct: return objPath[ct:]
            # now we have to see if an override is masking the path.
            tO=obj
            tI=IHostOverride.IHostOverride.isImplementedBy
            while hasattr(tO, "aq_parent"): # make the assumption that if there is already an alias in the path, the aliases must already be taken into account: fire away
                if tI(tO): return objPath[ct:]
                tO=tO.aq_parent
            fakeStack=list(objPath[ct:])
            fakePath="/" + "/".join(fakeStack)
            fakeStack.reverse()
            for over in overrides:
                res= over.testActivate(domainHost, fakeStack, fakePath)
                if res:
                    break # it's being masked by an override after all--we need to test the aliases for a match
            else: return objPath[ct:]
        # oof, here we go...we have to ask the alias overrides if they can provide a path to this resource...
        objURL="/"+"/".join(objPath)
        for over2 in overrides:
            if IHostOverride.IAliasHostOverride.isImplementedBy(over2):
                path=over2.getVirtualEquivalentPath(domainHost, objPath, objURL)
                if path: return path
        if res and raiseRedirect and IHostOverride.ISimpleRedirectHostOverride.isImplementedBy(res): # it's a redirect and we are supposed to raise it
            over.activate(res)
        if not res: # in other words, it's not that we've been masked by an override...
            try: # might as well see if acquisition will do the trick...
                testURL=self.getCanonicalVirtualRootURL(REQUEST)+ "/"+"/".join(objPath[ct:])
                if obj==self.unrestrictedTraverse(testURL):
                    return testURL.split("/")[1:]
            except:
                return None
        return None

manage_addHostForm=DTMLFile('dtml/addHost', globals())

def manage_addHost(dispatcher, id, REQUEST=None, title="", **ignored):
    """Constructor for Host"""
    container=dispatcher.Destination()
    obj = Host()
    obj._setId(str(id))
    if title: obj.title=title
    container._setObject(id, obj)
    if REQUEST:
        return dispatcher.manage_main(dispatcher,REQUEST, update_menu=1)
    else:
        return 1

constructors = (
  ('manage_addHostForm', DTMLFile('www/addHost', globals())),
  ('manage_addHost', manage_addHost),
)

InitializeClass(Host)
